/*
	Project Hamac
	Copyright 2004/5 D Henderson dh219@aibs.org.uk

	Contributors:
	David Rowbotham
	Toni Ylisirnio
	Russell Smith

	Licenced under the GNU General Public Licence (GPL)
	http://www.gnu.org/licenses/gpl.txt

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
#include <sys/types.h>

#ifdef _WIN32

#ifndef _WINSOCK2API_
#include <Winsock2.h>
#endif

#include <windows.h>
#include <mmsystem.h>
#include <vector>
#define snprintf _snprintf

#else

#include <sys/socket.h>
#include <vector.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <time.h>

#endif

#include <cmath>

#ifdef ORBITER_MODULE
#include "orbitersdk.h"
#else
#include "kosmos_math.h"
#endif


#define WINX1 -5.0*6372795.0
#define WINX2 5.0*6372795.0
#define WINY1 -5.0*6372795.0
#define WINY2 5.0*6372795.0
#define WINZ1 -5.0*6372795.0
#define WINZ2 5.0*6372795.0

#define D2R PI/180.0
#define R2D 180.0/PI
#define AU 1.4959787e+11

#define ANGLE3 VECTOR3

#define MAXUSERS 32
#define MAXPLANETS 32

#define TIMEACCEL 2.0

void rotatez( VECTOR3 *input, double head ) {

	VECTOR3 tmp;

	tmp.x = input->x * cos( head ) + input->y * sin ( head );
	tmp.y = input->y * cos( head ) - input->x * sin ( head );

	input->x = tmp.x;
	input->y = tmp.y;
}


class HVESSEL;

class USER {
public:
	USER( int filedesc, char *ipaddr, struct sockaddr_in *sockaddr );
//	USER( int filedesc, char *ipaddr );
	~USER();
	
	int fd;
	double lastgen;
	double lastship;
	char ip[20];
	struct sockaddr_in add;
	short udpport;
	double ping;
	unsigned int userid;

	double netfreq;
	
	HVESSEL *v;

	void sendnetworkdata();

	void InsertLag(double lagtime);
	double GetLagAvg();


private:

	double lag[30];
	short lagno;

};

class OBJECT {
public:
	OBJECT ( char *objname, double objmass, double objradius, VECTOR3 *p, VECTOR3 *v );
	~OBJECT ();

	VECTOR3 pos;
	VECTOR3 vel;
	VECTOR3 uacc;
	ANGLE3 arot;
	double mass;
	double size;
	char name[32];
	short type;
	bool drawn;
	
protected:

private:
	VECTOR3 gacc;
	VECTOR3 acc;
	
};

class BODY: public OBJECT {
public:
	BODY ( char *name, double mass, double objradius, VECTOR3 *p, VECTOR3 *v );
	~BODY ();
};
	
class HVESSEL: public OBJECT {
public:
	HVESSEL( char *name, double mass, double objradius, VECTOR3 *p, VECTOR3 *v, USER *user, unsigned int numeral );
	~HVESSEL();
	
	USER *u;
	unsigned int num;
	double lastupd;
	bool engine;
	bool laser;
	double lasertime;
	double ping;

	double lastdock;
	unsigned int dockedto;
	VECTOR3 dockpos;
	VECTOR3 dockarot;

#ifdef ORBITER_MODULE
	VESSEL *v;
#endif

};

std::vector<OBJECT*>obj;
std::vector<BODY*>bod;
std::vector<HVESSEL*>ves;
std::vector<USER*>use;



OBJECT::OBJECT ( char *objname, double objmass, double objradius, VECTOR3 *p, VECTOR3 *v ){
	pos=*p;
	vel=*v;
	
	mass=objmass;
	size=objradius;
	
	strncpy( name, objname, 31 );
	name[31] = 0;	
	
	acc = _V(0.0,0.0,0.0);
	gacc = _V(0.0,0.0,0.0);
	uacc = _V(0.0,0.0,0.0);
	arot = _V(0.0,0.0,0.0);

	obj.push_back(this);
}

OBJECT::~OBJECT () {
	std::vector<OBJECT*>::iterator i;
	for( i = obj.begin() ; i < obj.end() ; i++ ) {
		if( *i == this ) {
			obj.erase(i);
			break;
		}
	}
}

BODY::BODY ( char *objname, double objmass, double objradius, VECTOR3 *p, VECTOR3 *v ): OBJECT( objname, objmass, objradius, p, v ) {
	bod.push_back(this);
	type = 0;
}

BODY::~BODY() {
	std::vector<BODY*>::iterator i;
	for( i = bod.begin() ; i != bod.end() ; i++ ) {
		if( *i == this ) {
			bod.erase(i);
			break;
		}
	}
	
}

HVESSEL::HVESSEL( char *objname, double objmass, double objradius, VECTOR3 *p, VECTOR3 *v, USER *user, unsigned int numeral ): OBJECT( objname, objmass, objradius, p, v ) {
	ves.push_back(this);
	u = user;
	num = numeral;
	lastupd = 0.0;
	engine = false;
	laser = false;
	lasertime = 0.0;
	ping = 0.0;

	type = 1;

	lastdock = -999.0;
	dockedto = 0;
	dockpos=_V(0.0,0.0,0.0);
	dockarot=_V(0.0,0.0,0.0);

}

HVESSEL::~HVESSEL() {
	std::vector<HVESSEL*>::iterator i;
	for( i = ves.begin() ; i < ves.end() ; i++ ) {
		if( *i == this ) {
			ves.erase(i);
			break;
		}
	}	
}


USER *GetUserFromNum( int num ) {

	std::vector<USER*>::iterator i;

	for( i = use.begin() ; i < use.end() ; i++ ) {
		if( (*i)->v->num == num ) {
			return (*i);
		}
	}

	return NULL;
}

HVESSEL *GetVesselFromNum( int num ) {

	std::vector<HVESSEL*>::iterator i;

	for( i = ves.begin() ; i < ves.end() ; i++ ) {
		if( (*i)->num == num ) {
			return (*i);
		}
	}

	return NULL;
}

#ifdef ORBITER_MODULE
HVESSEL *GetVesselFromPtr( VESSEL *v ) {

	std::vector<HVESSEL*>::iterator i;

	for( i = ves.begin() ; i < ves.end() ; i++ ) {
		if( (*i)->v == v ) {
			return (*i);
		}
	}

	return NULL;
}
#endif

/* dh219 */
double mod( VECTOR3 *v ) {
	return( sqrt( v->x*v->x + v->y*v->y + v->z*v->z ) );
}

double mod( VECTOR3 v ) {
	return( sqrt( v.x*v.x + v.y*v.y + v.z*v.z ) );
}


void vecrot( VECTOR3 *in, ANGLE3 *arot, VECTOR3 *out ) {

	VECTOR3 tmp, tmp2;
	double a,b,c;

	a = arot->x;
	b = arot->y;
	c = arot->z;

	tmp.x = in->x;
	tmp.y = cos(a)*in->y + sin(a)*in->z;
	tmp.z = cos(a)*in->z - sin(a)*in->y;

	tmp2.x = cos(b)*tmp.x - sin(b)*tmp.z;
	tmp2.y = tmp.y;
	tmp2.z = sin(b)*tmp.x + cos(b)*tmp.z;

	out->x = cos(c)*tmp2.x - sin(c)*tmp2.y;
	out->y = cos(c)*tmp2.y + sin(c)*tmp2.x;
	out->z = tmp2.z;

}

void vecunrot( VECTOR3 *in, ANGLE3 *arot, VECTOR3 *out ) {

	VECTOR3 tmp, tmp2;
	double a,b,c;

	a = -arot->x;
	b = -arot->y;
	c = -arot->z;



	tmp.x = cos(c)*in->x - sin(c)*in->y;
	tmp.y = cos(c)*in->y + sin(c)*in->x;
	tmp.z = in->z;

	tmp2.x = cos(b)*tmp.x - sin(b)*tmp.z;
	tmp2.y = tmp.y;
	tmp2.z = sin(b)*tmp.x + cos(b)*tmp.z;

	out->x = tmp2.x;
	out->y = cos(a)*in->y + sin(a)*in->z;
	out->z = cos(a)*in->z - sin(a)*in->y;

}
