/*
	Project Hamac
	Copyright 2004/5 D Henderson dh219@aibs.org.uk

	Contributors:
	David Rowbotham
	Toni Ylisirnio
	Russell Smith

	Licenced under the GNU General Public Licence (GPL)
	http://www.gnu.org/licenses/gpl.txt

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


//	#define DEBUG

#define STRICT
#define ORBITER_MODULE


#define GM 4.9189533e12 // moon

#define MAXDOCK 10

#include "/orbits/orbits/defines.h"

char version[128];

DWORD start;
double realtime;
double nettime;
double netrectime;
double oldnetrectime;
double dnrt;
double nt, oldnt;
double dnt;
double dt;
double oldt;
double netping;

double lastkeytime;

unsigned int userid;

VECTOR3 olduacc;
VECTOR3 oldarot;
//VECTOR3 localacc;
VECTOR3 zero;

#define GMmoon 398600.4418*1e9 // moon

HVESSEL *me;

int fdm;
int fdu;
WSAData WSAdat;

struct sockaddr_in dest;

long _PORT;
//char *_HOSTNAME="192.168.2.2";
//char *_HOSTNAME="10.168.131.54";
char *_HOSTNAME="ipmon.aibs.org.uk";
//char *_HOSTNAME="wcklinnas01-pool0-a3.wcklin.tds.net";
//char *_HOSTNAME="anakin.mavericktech.com.au";

OBJHANDLE moon;

char fp[512];
FILE *fpd;
bool active;

struct {
	HVESSEL *dockedto;
	HVESSEL *wasdockedto;
	double undocktime;
} docks[MAXDOCK];

/********************************************************************************************/

void udping() {
	struct sockaddr_in tmp;
	tmp = dest;
	tmp.sin_port = htons(_PORT);
	memset(&(tmp.sin_zero),0,8);
	char buf[512];
	if( userid != 999 && fdu ) {
		_snprintf( buf, sizeof( buf )-1, "I %d %f PING\n", userid, netrectime );
		sendto( fdu, buf, strlen(buf), 0, (struct sockaddr*)&tmp, sizeof( struct sockaddr ) );
	}
}

void startnetworking() {

	struct sockaddr_in myadd;
	struct hostent *host;

	WSAStartup(0x0fff,&WSAdat);

	host = gethostbyname( _HOSTNAME );
	if( host == NULL ) {
		sprintf( fp, "Couldn't get %s\n", _HOSTNAME );
		MessageBox( NULL,fp,"Hamac: Error",MB_ICONINFORMATION);
		exit(2);
	}

	fdm = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP );
	if( fdm == -1 ) {
		sprintf( fp, "Socket init error\n" );
		MessageBox( NULL,fp,"Hamac: Error",MB_ICONINFORMATION);
		exit(1);
	}
	fdu = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );
	if( fdu == -1 ) {
		sprintf( fp, "Socket init error\n" );
		MessageBox( NULL,fp,"Hamac: Error",MB_ICONINFORMATION);
		exit(1);
	}

	myadd.sin_family = AF_INET;
	myadd.sin_port = htons(0);
	myadd.sin_addr.s_addr = htonl(0);
	memset(&(myadd.sin_zero),0,8);
	if( bind( fdu, (struct sockaddr*)&myadd, sizeof( struct sockaddr ) ) ) {
		sprintf( fp, "Socket bind error\n" );
		MessageBox( NULL,fp,"Hamac: Error",MB_ICONINFORMATION);
		exit(1);
	}

	int l = sizeof( struct sockaddr );
	getsockname( fdu, (struct sockaddr*)&myadd, &l ) ;
//	printf("localudpport: '%d'\n", ntohs(myadd.sin_port) );
	

	dest.sin_family = AF_INET;
	dest.sin_port = htons(_PORT);
	dest.sin_addr = *((struct in_addr*)host->h_addr);
	memset(&(dest.sin_zero),0,8);

	if( connect( fdm, (struct sockaddr*)&dest, sizeof( struct sockaddr ) ) ) {
		sprintf( fp, "Socket connect error\n" );
		MessageBox( NULL,fp,"Hamac: Error",MB_ICONINFORMATION);
		exit(1);
	}

}

void nsend(int *fdl, char *buf, unsigned int len ) {
	int bytes;

	bytes = send( *fdl, buf, len, 0 );
	if( bytes < 0 ) {
		shutdown( *fdl, 2 );
		fprintf( stderr, "Shutting down stream\n" );
		exit(3);
	}
}

void senddock() {

	VESSEL *v, *o;
	HVESSEL *ho;
	VESSELSTATUS vs, vs2;
	unsigned int i;
	char buf[1024];

	v = me->v;
	v->GetStatus(vs);

	for( i = 0 ; i < MAXDOCK && i < v->DockCount() ; i++ ) {
		if( v->DockingStatus(i) ) {
			o = oapiGetVesselInterface(v->GetDockStatus(v->GetDockHandle(i)));
			docks[i].dockedto = GetVesselFromPtr(o);
			docks[i].wasdockedto = docks[i].dockedto;
			docks[i].undocktime = -1.0;
		}
		else if( docks[i].dockedto ) {
			docks[i].dockedto = NULL;
			docks[i].undocktime = realtime;
			if( docks[i].wasdockedto ) {
				_snprintf( buf, sizeof( buf )-1 , "U %d\n", docks[i].wasdockedto->num ); 
				if( fdm )
					nsend( &fdm, buf, strlen( buf ) );
				return;
			}

		}

	}
	
	for( i = 0 ; i < MAXDOCK ; i++ ) {
		if( ( ho = docks[i].dockedto ) ) {
			VECTOR3 relpos, lrelpos;
			o->GetStatus(vs2);
			relpos = vs2.rpos-vs.rpos;
//			vecunrot( &relpos, &(vs.arot), &lrelpos );
			lrelpos = relpos;

			_snprintf( buf, sizeof( buf )-1 , "D %d %f %f %f %f %f %f\n",
				ho->num, lrelpos.x, lrelpos.y, lrelpos.z ,
				vs2.arot.x, vs2.arot.y, vs2.arot.z );
			sprintf( oapiDebugString(), "%s", buf );
			if( fdm )
				nsend( &fdm, buf, strlen( buf ) );
		}
	}
}

void sendacc() {

	char buf[2048];
	VECTOR3 arot;
	VECTOR3 uacc;
	VECTOR3 lacc;
	VESSEL *v;
	THRUSTER_HANDLE th;

	VECTOR3 ftmp, Ttmp;
	VECTOR3 f, T;
	VECTOR3 pmi;


	v = me->v;

	if( !v )
		return;

	v->GetPMI(pmi);

	int i;

	v = me->v;

	f = T = zero;

	for( i = 0 ; i < v->GetThrusterCount() ; i++ ) {
		th = v->GetThrusterHandleByIndex(i);
		v->GetThrusterMoment( th, ftmp, Ttmp );

		f += ftmp;
		T += Ttmp;

	}

	lacc.x = f.x/v->GetMass();
	lacc.y = f.y/v->GetMass();
	lacc.z = f.z/v->GetMass();

	v->GlobalRot( lacc, uacc );

	VESSELSTATUS vs;
	v->GetStatus(vs);

	arot = vs.arot;


	VECTOR3 diff, diffr;
	
	diff = uacc - olduacc;
	diffr = arot - oldarot;



	if( fabs( mod(diff) ) > 0.01 || fabs( mod(diffr) ) > 0.0001 ) {
		olduacc = uacc;
		oldarot = arot;
		_snprintf( buf, sizeof( buf )-1, "A %f %f %f %f %f %f %f\n", nettime, uacc.x, uacc.y, uacc.z, arot.x, arot.y, arot.z  );
		if( fdm )
			nsend( &fdm, buf, strlen(buf) );
	}


}


bool clientlisten() {

	unsigned int sin_size;
	fd_set r;
	fd_set w;
	fd_set e;
	struct timeval timeout;
	struct sockaddr_in theiradd;

	char buf[4096];
	VECTOR3 p,v, arot;
	unsigned int vesno, rel;

	int rb;
	bool retval;

	sin_size = sizeof( struct sockaddr_in );
	
	FD_ZERO( &r );
	FD_ZERO( &w );
	FD_ZERO( &e );
	timeout.tv_sec = 0;
	timeout.tv_usec = 10;

	int max;
	max = -1;
	if( fdu ) {
		FD_SET( fdu, &r );
		max = fdu;
	}
	if( fdm ) {
		FD_SET( fdm, &r );
		if( fdm > max )
			max = fdm;
	}

	retval = false;

	select( max+1, &r, NULL, NULL, &timeout );

	if( FD_ISSET( fdm, &r ) ) {
		rb = recv( fdm, buf, sizeof( buf  )-1, 0 ) ;
		buf[rb] = 0;
		if( rb < 0 ) {
			sprintf( fp, "Closing connection\n");
			MessageBox( NULL,fp,"Hamac: Error",MB_ICONINFORMATION);
			shutdown( fdu, 2 );
			fdu = -1;
			shutdown( fdm, 2 );
			fdm = -1;
		}
		else {
			if( strlen(buf) > 2 && buf[0] == 'I' ) {
				userid = (unsigned int)atoi(buf+2);
				me->num = userid;
				udping();
			}
		}
		buf[0] = 0;
	}

	if( FD_ISSET( fdu, &r ) ) {
		rb = recvfrom( fdu, buf, sizeof( buf  )-1, 0, (struct sockaddr*)&theiradd, (int*)&sin_size ) ;
		buf[rb] = 0;

#ifdef DEBUG
//		fprintf( fpd, "%d (%s)\n\n", rb, buf );
#endif

//		sprintf( oapiDebugString(), "r: '%s'", buf );
		
		if( rb < 0 ) {
			sprintf( fp, "Closing connection\n");
			MessageBox( NULL,fp,"Hamac: Error",MB_ICONINFORMATION);
			shutdown( fdu, 2 );
			fdu = -1;
			shutdown( fdm, 2 );
			fdm = -1;
		}

		char buf2[4096];
		char *ptr;


		while( ( ptr = strstr( buf, "\n" ) ) ) {
			strncpy( buf2, buf, ptr-buf );
			buf2[ptr-buf] = 0;

			strcpy( buf, ptr+1 );
#ifdef DEBUG
			fprintf( fpd, "bufleft: '%s'\n\n", buf );
#endif
	
			/* undock! */

			if( strlen(buf2) > 2 && buf2[0] == 'U' ) {
				unsigned int s;
				s = (unsigned int)atoi(buf2+2);
				if( s == 0 )
					break;

				VESSEL *v;
				v=me->v;

				int i;
				for( i = 0 ; i < v->DockCount() && i < MAXDOCK ; i++ ) {
					if( docks[i].wasdockedto == GetVesselFromNum(s) ) {
						v->Undock(i,0);
//						sprintf(oapiDebugString(), "Undocking %.2f", realtime );
						break;
					}
				}
			}




			/* vessel co-ords rec'd */
			if( strlen(buf2) && buf2[0] == 'V' ) {

				sscanf(buf2+2, "%i %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %i",
					&vesno, &netrectime, &p.x, &p.y, &p.z, &v.x, &v.y, &v.z,
					&arot.x, &arot.y, &arot.z, &rel );

#ifdef DEBUG
				fprintf( fpd ,"%ld: '%s'\n", timeGetTime(), buf2 );
#endif


				if( nettime < 0.0 )
					nettime = netrectime;

				/* check we have the latest packet */
//				if( netrectime > oldnetrectime-(0.01) ) {
				{

					if( vesno == 0 )
						udping();

					/* wot vessel? */

					HVESSEL *vv;

					vv = NULL;

					if( vesno == 0 )
						vv = me;
					else {
						vv = GetVesselFromNum(vesno);
					}

					if( vv == NULL ) {
						/* new vessel */

						VESSELSTATUS vs;
						moon = oapiGetObjectByName("Moon");

						vs.rpos=_V(0.0,1e7,0.0);
						vs.rvel=_V(GMmoon/1e7,0.0,0.0);
						vs.vrot = zero;
						vs.arot=zero;
						vs.fuel = 1.0;
						vs.eng_hovr = 0.0;
						vs.eng_main = 0.0;
						vs.rbody = moon;
						vs.status = 0;

						char buf[128];
						_snprintf(buf, sizeof( buf )-1, "R-%2.2d", vesno );
						OBJHANDLE ome;
						ome = oapiCreateVessel(buf,"deltaglider",vs);
						VESSEL *v;
						v = oapiGetVesselInterface(ome);
						
						HVESSEL *vvv;
						vvv = new HVESSEL( buf, v->GetMass(), v->GetSize(), &vs.rpos, &vs.rvel, NULL, 0 );
						vvv->v = v;
						vvv->ping = realtime;
						vvv->num = vesno;
						retval=true;


/*						while( v->DockCount() > 0 ) {
							v->DelDock(v->GetDockHandle(0));
						}
*/

						return true;
					} 
					else if( netrectime > vv->lastupd ) {
						VESSELSTATUS status;
						vv->v->GetStatus(status);

						printf("%d\n", vesno );

						moon = oapiGetObjectByName("Moon");

						/* update vectors */
						vv->pos.x = p.x;
						vv->pos.y = p.y;
						vv->pos.z = p.z;

						vv->vel.x = v.x;
						vv->vel.y = v.y;
						vv->vel.z = v.z;

						vv->arot.z = arot.z;


						VECTOR3 ptmp, vtmp;
						ptmp.x = p.x;
						ptmp.y = p.y;
						ptmp.z = p.z;

						vtmp.x = v.x;
						vtmp.y = v.y;
						vtmp.z = v.z;

						status.rpos = ptmp;
						status.rvel = vtmp;
						status.rbody = moon;

						if( me != vv ) {
							status.arot = arot;
							status.vrot = zero;
						}


						dnrt = nettime - netrectime;
						oldnetrectime = netrectime;
						vv->lastupd = netrectime;


						vv->ping = realtime;
				

						vv->v->DefSetState(&status);

						retval=true;
					}
				}
			}

		}

	}
	return retval;
}

void undockcheck() {
	

}


/************* ORBITER BITS *************************************************/

DLLCLBK void opcOpenRenderViewport( HWND renderWnd, DWORD width, DWORD height, BOOL fullscreen) {

	sprintf(oapiDebugString(),  "%s", version );
	
	moon = oapiGetObjectByName("Moon");
	
	VESSELSTATUS vs;

	vs.rpos=_V(0.0,1e7,0.0);
	vs.rvel=_V(GMmoon/1e7,0.0,0.0);
	vs.vrot = zero;
	vs.arot=zero;
	vs.fuel = 1.0;
	vs.eng_hovr = 0.0;
	vs.eng_main = 0.0;
	vs.rbody = moon;
	vs.status = 0;

	ves.clear();

	OBJHANDLE ome;
	ome = oapiCreateVessel("User","deltaglider",vs);
	VESSEL *v;
	v = oapiGetVesselInterface(ome);
	
	me = new HVESSEL( "User", v->GetMass(), v->GetSize(), &vs.rpos, &vs.rvel, NULL, 0 );
	me->v = v;
	me->ping = 0.0;
	me->num = 0;

	nettime = -1.0;
	netping = 0.0;
	
	start = timeGetTime();
	oldnetrectime = 0.0;
	netrectime = -1.0;

	lastkeytime = 0.0;

	userid = 999;

	startnetworking();
  
/* ship stuff */
	realtime = 0.0;
	oldt = 0.0;
	dt = 0.0;

	oapiSetTimeAcceleration( TIMEACCEL );

	active = true;

	int j;
	for( j = 0 ; j < MAXDOCK ; j++ ) {
		docks[j].dockedto = NULL;
		docks[j].wasdockedto = NULL;
		docks[j].undocktime = -1.0;
	}

			
	oapiSetFocusObject(ome);
}

DLLCLBK void opcCloseRenderViewport (void) {
	active = false;

	while ( !ves.empty()  ) {
		delete ves[0];
	}

	if( fdm > -1 ) {
		shutdown( fdm, 2 );
		fdm = -1;
	}

	if( fdu > -1 ) {
		shutdown(fdu,2);
		fdu = -1;
	}

}

DLLCLBK void opcDLLInit (HINSTANCE hDLL)
{

	sprintf( version, "Project Hamac. (C) 2005 D Henderson based on works by Toni Ylisirnio (IRCMFD). Build %s %s", __DATE__, __TIME__ );

	active = false;
	me = NULL;

#ifdef DEBUG
	fpd = fopen("hamac.dbg", "w" );
#endif

	
	_PORT = 1977L;

	zero = _V(0.0,0.0,0.0);

	olduacc=zero;
	oldarot=zero;

}


DLLCLBK void opcDLLExit (HINSTANCE hDLL)
{
#ifdef DEBUG
	fclose(fpd);
#endif
}

DLLCLBK void opcTimestep (double simt, double simdt, double mjd)
{

	if( !active )
		return;

  /* Increase time  */  

	realtime = (timeGetTime() - start) / 1000.0;
	dt = realtime - oldt;

	if( dt < 0.05 ) return;

	oldt = realtime;
	if( nettime > 0.0 )
		nettime += dt;

	senddock();
	undockcheck();
	sendacc();

	clientlisten();


	/* clean up missing ships -- always seems to crash! */
	std::vector<HVESSEL*>::iterator i;
	for( i = ves.begin() ; i < ves.end() ; i++ ) {
		
		if( (*i)->ping + 10.0 < realtime && (*i)->ping > 0.000001 ) {
			try {
				int j;
				for( j = 0 ; j < MAXDOCK ; j++ ) {
					if( docks[j].dockedto == (*i) ) {
						docks[j].dockedto = NULL;
						docks[j].wasdockedto = NULL;
						docks[j].undocktime = -1.0;
					}
				}
				oapiDeleteVessel((*i)->v->GetHandle());
				if( (*i) == me ) {
					me = NULL;
					active = false;
					sprintf(fp,"Lost connection to server, shutting down.");
					MessageBox( NULL,fp,"Hamac: Error",MB_ICONINFORMATION);
					if( fdm != -1 )
						shutdown( fdm, 2 );
					if( fdu != -1 )
						shutdown( fdu, 2 );
				}
				delete (*i);
				return;
			}
			catch( ... ) { 
				sprintf(fp,"Couldn't delete vessel!");
				MessageBox( NULL,fp,"Hamac: Error",MB_ICONINFORMATION);
			}
		}
	}



#ifdef DEBUG
	std::vector<HVESSEL*>::iterator hv;
	VECTOR3 pos;
	for( hv = ves.begin() ; hv < ves.end() ; hv++ ) {
		if( (*hv)->v ) {
			oapiGetRelativePos( (*hv)->v->GetHandle(), moon, &pos );
			fprintf( fpd, "%.2f: %d - %.2f %.2f %.2f\n", realtime, (*hv)->num, pos.x, pos.y, pos.z );
		}
	}
#endif


#ifdef DEBUG
	fflush(fpd);
#endif
}
