/*
	Project Hamac
	Copyright 2004/5 D Henderson dh219@aibs.org.uk

	Contributors:
	David Rowbotham
	Toni Ylisirnio
	Russell Smith

	Licenced under the GNU General Public Licence (GPL)
	http://www.gnu.org/licenses/gpl.txt

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*
* Kosmos
* Authors :
* Dave Rowbotham ( Kosmos Architecture & Programming )
* Russell Smith  ( API Architecture & Programming )
* John Bowles ( Unix Port & Programming )
* Sean O'Neil. ( CVertex, CVertexArray, Also the
*   ROAM Engine is based around Seans published LGPL'd ROAM Code )
*
* Major Contributors.
* David Henderson ( Analytical and numeric testing as well as Physics consulting )
* Toni Ylisirni   ( OpenGL and Allegro consulting and Shader code )
* Simon Proud   ( Space Systems and Physics consulting )
* Duncan Sharpe ( for help with Runge Kutta algorithm and for the TransX project )
* Dorian Pascoe ( for help with Runge Kutta algorithm )
* Martin Schweiger ( for inspiring us with Orbiter and its API )
*/

#include "kosmos_math.h"


/**
* @author Dave Rowbotham.
* @param double x, double y, double z
*/
VECTOR3 _V( double x, double y, double z ){
  return VECTOR3(x,y,z);
}


/**
* @author Dave Rowbotham.
* @param double nx1, double ny1, double nz1
* Implementation of Orbiter Compatible Class VECTOR3
*/
VECTOR3::VECTOR3 (double nx1, double ny1, double nz1){
  x = nx1;
  y = ny1;
  z = nz1;
  ///< Orbiter Compatability
  data[0] = x;
  data[1] = y;
  data[2] = z;
}

/**
* @author Dave Rowbotham.
* @param const VECTOR3 & other
*/
VECTOR3::VECTOR3 (const VECTOR3 & other){
  x = other.x;
  y = other.y;
  z = other.z;
  nx=other.nx;
  ny=other.ny;
  nz=other.nz;
  tu=other.tu;
  tv=other.tv;
  ///< Orbiter Compatability
  data[0] = x;
  data[1] = y;
  data[2] = z;
}

/**
* @author Dave Rowbotham.
* @param int i
*/
double & VECTOR3::operator[] (int i){
  static double _failed=0.0;
  switch(i) {
    case 0: return x;
    case 1: return y;
    case 2: return z;
  }
  return _failed;
}

/**
* @author Dave Rowbotham.
* @param const VECTOR3 & other
*/
VECTOR3 & VECTOR3::operator= (const VECTOR3 & other){
  x = other.x;
  y = other.y;
  z = other.z;
  nx=other.nx;
  ny=other.ny;
  nz=other.nz;
  tu=other.tu;
  tv=other.tv;
  ///< Orbiter Compatability
  data[0] = x;
  data[1] = y;
  data[2] = z;
  return *this;
}

/**
* @author Dave Rowbotham.
* @param const VECTOR3 & other
*/
VECTOR3 & VECTOR3::operator+=(const VECTOR3 & other){
  x += other.x;
  y += other.y;
  z += other.z;
  ///< Orbiter Compatability
  data[0] = x;
  data[1] = y;
  data[2] = z;
  return *this;
}

/**
* @author Dave Rowbotham.
* @param const VECTOR3 & other
*/
VECTOR3 VECTOR3::operator+(const VECTOR3 & other) const{
  VECTOR3 vv;
  vv.x = x + other.x;
  vv.y = y + other.y;
  vv.z = z + other.z;
  return vv;
}

/**
* @author Dave Rowbotham.
* @param const VECTOR3 & other
*/
VECTOR3 VECTOR3::operator-(const VECTOR3 & other) const{
  VECTOR3 vv;
  vv.x = x - other.x;
  vv.y = y - other.y;
  vv.z = z - other.z;
  return vv;
}

/**
* @author Dave Rowbotham.
* @param
*/
VECTOR3 VECTOR3::operator- (){
  return VECTOR3 (-x, -y, -z);
}

/**
* @author Dave Rowbotham.
* @param double scalar
*/
VECTOR3 VECTOR3::operator* (double scalar) const{
  return VECTOR3 (x * scalar, y * scalar, z * scalar);
}

/**
* @author Dave Rowbotham.
* @param double scalar
*/
VECTOR3 VECTOR3::operator/ (double scalar) const{
  return VECTOR3(x / scalar, y / scalar, z / scalar);
}

/**
* @author Dave Rowbotham.
* @param  double radius, double latitude, double longitude
*/
VECTOR3 VECTOR3::PolarToCartesian( double radius, double latitude, double longitude ) {
  VECTOR3 result;
  result.z = (radius) * cos(latitude * PI / 180) * cos(longitude * PI / 180);
  result.y = (radius) * sin(latitude * PI / 180);
  result.x = (radius) * cos(latitude * PI / 180) * sin(longitude * PI / 180);
  return result;
}

/**
* @author Dave Rowbotham.
* @param
*/
double VECTOR3::Length (){
  return sqrt (x*x+y*y+z*z);
}

/**
* @author Dave Rowbotham.
* @param
*/
VECTOR3 VECTOR3::Unit (){
  return (*this) / Length();
}

/**
* @author Dave Rowbotham.
* @param
*/
double VECTOR3::Normalise() {
  double len = Length();
  if (fabs(len)<1e-15) {
    x =	0.0f;
    y =	0.0f;
    z =	0.0f;
  }  else {
    x = (double)(x / len);
    y = (double)(y / len);
    z = (double)(z / len);
  }
  return(len);
}
