/*
	Project Hamac
	Copyright 2004/5 D Henderson dh219@aibs.org.uk

	Contributors:
	David Rowbotham
	Toni Ylisirnio
	Russell Smith

	Licenced under the GNU General Public Licence (GPL)
	http://www.gnu.org/licenses/gpl.txt

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef _KOSMOS_MATHS_H_
#define _KOSMOS_MATHS_H_
#endif
///<  a suitably intimidating value for PI
#ifndef PI
  #define PI 3.141592653589793238462643383279502884
  #define TWOPI 2.0*PI
  #define PID2  PI/2.0
#endif

#include <cmath>
#include <iostream>
#define LOGHALF		-0.6931471805599f
#define LOGHALFI	-1.442695040889f
#define DELTA		1e-6f
#define MAX_DIMENSIONS	4
#define MAX_OCTAVES	128
#define DEGTORAD(x)	((x) * 0.01745329251994f)
#define RADTODEG(x)	((x) * 57.29577951308f)
#define SWAP(a, b, t)	t = a; a = b; b = t

/**
*  Orbiter Compatible 3d Vector Class
*  Also provides some extra functionality like stream IO
*  and GLFloat pointers
*/
class VECTOR3{
public:
  VECTOR3 PolarToCartesian( double radius, double latitude, double longitude );
  VECTOR3 (double x=0, double y=0, double z=0);
  VECTOR3 (const VECTOR3 & other);
  double &operator[] (int i);			///<  Subscript
  VECTOR3 &operator= (const VECTOR3 & other);	///<  Assign
  VECTOR3 &operator+= (const VECTOR3 & other);	///<  Add and assign
  bool	operator==( const VECTOR3 & other) const;
  VECTOR3 operator+ (const VECTOR3 & other) const;///<  Vector addition
  VECTOR3 operator- (const VECTOR3 & other) const;///<  Vector difference
  VECTOR3 operator- ();			  ///<  Vector negation
  VECTOR3 operator* (double scalar) const;///<  Vector * Scalar
  VECTOR3 operator/ (double scalar) const;///<  Vector / Scalar
  VECTOR3 Unit ();	///<  Unit vector in this direction
  double Length ();	///<  Vector Length
  int 	IsUnit();
  double Normalise();
  void 	ZeroClamp();
  int 	IsZero();
  ///<  da data!.
  double x,y,z,nx,ny,nz,tu,tv;
  ///< Orbiter compatability
  double data[3];
};
class POINT3 : public VECTOR3 {};
extern VECTOR3 _V( double x, double y, double z );

///<  convert 3 double parameters to VECTOR3
VECTOR3 _V(double x,double y, double z);


