/*
	Project Hamac
	Copyright 2004/5 D Henderson dh219@aibs.org.uk

	Contributors:
	David Rowbotham
	Toni Ylisirnio
	Russell Smith

	Licenced under the GNU General Public Licence (GPL)
	http://www.gnu.org/licenses/gpl.txt

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



/* central potential generator GM in m^3/s^2 */

//#define GM 398600.4418*1e9 // earth
#define GM 4.9189533e12 // moon
#define RADIUS 1.738e6

#ifdef _WIN32
#define snprintf _snprintf
#endif

#include "orbits.h"

#define startpos _V( 2e6, 0.0, 0.0 )

#define LINES 1000

#define STEPMIN 1e-3
#define STEPMAX 120

#define REPORTSTEP 0.3

#define MAXUSERS 32

#define STDIN 0

#define LASERTIME 0.2

double start;
double dt;
double oldt;
double t;


struct hostent *he;
int fdm; // main file descriptor
int fdu; // main udp descriptor

//int fdc[MAXUSERS];

/* END OF HEADERS */

long _PORT;

/*#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
*/

/*
void resetships() {

	pos = startpos;
	vel.x = 0.0;	
	vel.y = sqrt( 1.0*GM/pos.x );	
	vel.z = 0.0;	
	o.v = _V( 0.0, 0.0, 0.0 );
	o.rot = 0.0;
	oldo.v = _V( 0.0, 0.0, 0.0 );
	oldo.rot = 0.0;
	
	vacc = _V(0.0,0.0,0.0);
	
}
*/

void startnetworking() {

struct sockaddr_in myaddr;

#ifdef _WIN32
	WSAData WSAdat;
	WSAStartup(0x0fff,&WSAdat);
	fdm = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP );
	fdu = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );
#else
	fdm = socket( AF_INET, SOCK_STREAM, /*IPPROTO_TCP*/0 );
	fdu = socket( AF_INET, SOCK_DGRAM, /*IPPROTO_TCP*/0 );
#endif

	if( fdm == -1 || fdu == -1 ) {
		fprintf( stderr, "Socket init error\n" );
		exit(1);
	}

	myaddr.sin_family = AF_INET;
	myaddr.sin_port = htons(_PORT);
	myaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	memset(&(myaddr.sin_zero),0,8);

	if( bind( fdm, (struct sockaddr*)&myaddr, sizeof( struct sockaddr) ) ) {
		fprintf( stderr, "Socket bind error\n");
		exit(1);
	}

	if( bind( fdu, (struct sockaddr*)&myaddr, sizeof( struct sockaddr) ) ) {
		fprintf( stderr, "Socket bind error\n");
		exit(1);
	}

	if( listen( fdm, 5 ) ) {
		fprintf( stderr, "Socket listen error\n" );
		exit(1);
	}

}
void nsend( USER *u, char *buf, unsigned int len ) {

	int bytes;
	int fdl;

	struct sockaddr_in theiradd;
	
	fdl = u->fd;
	
	if( fdl == -1 || u->udpport == -1 )
		return;
	
	theiradd=u->add;

	theiradd.sin_port=htons(u->udpport);
	theiradd.sin_family = AF_INET;
	memset(&(theiradd.sin_zero),0,8);


	bytes = sendto( fdu, buf, len, 0, (struct sockaddr*)&theiradd, sizeof(struct sockaddr) );
//	printf("Sent: %d bytes\n", bytes );


	if( bytes < 0 ) {
		fprintf( stderr, "Shutting down stream from %s\n", u->ip );
		shutdown(fdl,2);
		u->fd = -1;
		fdl = u->fd;
	}	
}

void serverlisten() {

	unsigned int sin_size;
	int newfd;
	struct sockaddr_in remaddr;
	fd_set r;
	fd_set w;
	fd_set e;
	struct timeval timeout;
	int max;
	int fdl;
	
	char buf[512];
	char buf2[512];
	char *ptr;

	sin_size = sizeof( struct sockaddr_in );
	
	FD_ZERO( &r );
	FD_ZERO( &w );
	FD_ZERO( &e );
	timeout.tv_sec = 0;
	timeout.tv_usec = 0;

#ifndef _WIN32
	FD_SET( STDIN, &r );
#endif
	FD_SET( fdm, &r );
	max = fdm;
	FD_SET( fdu, &r );
	if( fdu > max )
		max = fdu;

	std::vector<USER*>::iterator j;

	for( j = use.begin() ; j < use.end() ; j++ ) {
		fdl = (*j)->fd;
		if( fdl != -1 ) {
			FD_SET( fdl, &r );
			if( fdl > max )
				max = fdl;
		}
	}


	select( max+1, &r, NULL, NULL, &timeout );

	if( FD_ISSET( STDIN, &r ) ) {
		fgets( buf, sizeof( buf )-1, stdin );
		if( strlen( buf ) )
			if( buf[0] == 'q' ) {
				shutdown(fdm,2);
				shutdown(fdu,2);
				exit(0);
			}
	}


	if( FD_ISSET( fdm, &r ) ) {
		if( use.size() >= MAXUSERS ) {
			fprintf( stderr, "Too many users! Rejecting connection.\n" );
			return;
		}

		
#ifdef _WIN32
		newfd = accept( fdm, (struct sockaddr*)&remaddr, (int*)&sin_size );
#else
		newfd = accept( fdm, (struct sockaddr*)&remaddr, &sin_size );
#endif
		if( newfd == -1 ) {
			fprintf( stderr, "No accept necessary\n" );
			return;
		}
		
		USER *u;
		u =	new USER( newfd, inet_ntoa( remaddr.sin_addr ), &remaddr );

		if( u->userid == 999 ) {
			shutdown( newfd, 2 );
			delete u;
		}
		else {
			printf("Server connection from: %s. Sending userid %i\n", inet_ntoa( remaddr.sin_addr ), u->userid );
			char buf[512];
#ifdef _WIN32
			_snprintf( buf, sizeof( buf )-1, "I %d\n", u->userid );
#else
			snprintf( buf, sizeof( buf )-1, "I %d\n", u->userid );
#endif
			send( newfd, buf, strlen(buf), 0 );
		}
	}

	int rb;



	if( FD_ISSET( fdu, &r ) ) {
#ifdef _WIN32
		rb = recvfrom( fdu, buf, sizeof( buf  )-1, 0, (struct sockaddr*)&remaddr, (int*)&sin_size ) ;
#else
		rb = recvfrom( fdu, buf, sizeof( buf  )-1, 0, (struct sockaddr*)&remaddr, &sin_size ) ;		
#endif
		if( rb < 0 ) {
			fprintf( stderr, "Error reading UDP channel from %s\n", inet_ntoa( remaddr.sin_addr ) );
		}
		else {
			buf[rb] = 0;
//			printf("UDP(%s): '%s'\n", inet_ntoa(remaddr.sin_addr), buf );
			if( strlen( buf ) > 4 && buf[0] == 'I' ) {
				unsigned int id;
				double replytime;
				sscanf( buf, "I %d %lf", &id, &replytime );
//				printf("%s\n", buf );
				for( j = use.begin() ; j != use.end() ; j++ ) {
					if( (*j)->userid == id ) {
						(*j)->udpport = ntohs(remaddr.sin_port);
						(*j)->ping = t;
						(*j)->InsertLag(t-replytime);
//						printf("Ping %d:%.2f\n", (*j)->v->num, t );
						break;
					}
				}
			} 
			
		}
	}




	for( j = use.begin() ; j != use.end() ; j++ ) {

		fdl = (*j)->fd;
		if( fdl != -1 ) {
			if( FD_ISSET( fdl, &r ) ) {
				/* listen here */

				rb = recv( fdl, buf2, sizeof( buf2  )-1, 0 ) ;
				if( rb <= 0 ) {
					fprintf( stderr, "Closing connection from %s\n", (*j)->ip );
					shutdown(fdl,2);
					(*j)->fd = -1;
					fdl = (*j)->fd;
					break;
				}
				buf2[rb]=0;
//				printf("%s: '%s'\n", inet_ntoa((*j)->add.sin_addr),buf2);
				
				HVESSEL *v;
				v = (*j)->v;

				while( ( ptr = strstr( buf2, "\n" ) ) ) {
					strncpy( buf, buf2, ptr-buf2 );
					buf[ptr-buf2] = 0;

					strcpy( buf2, ptr+1 );
				
					
					if( buf[0] == 'A' ) {
						double nt;
						sscanf( buf+2, "%lf %lf %lf %lf %lf %lf %lf", &nt, &(v->uacc.x), &(v->uacc.y), &(v->uacc.z), &(v->arot.x), &(v->arot.y), &(v->arot.z) );
						if( strstr(buf+2, ":M:") )
							v->engine = true;
						else
							v->engine = false;
					}
					if( buf[0] == 'O' ) {
						sscanf( buf+2, "%lf %lf %lf", &(v->arot.x), &(v->arot.y), &(v->arot.z) );
					}
					if( buf[0] == 'D' && strlen( buf ) > 2 ) {
						VECTOR3 dockpos, arot;
						unsigned int dockedto;
						USER *u;

						sscanf( buf+2, "%d %lf %lf %lf %lf %lf %lf", &dockedto, &(dockpos.x), &(dockpos.y), &(dockpos.z), &(arot.x), &(arot.y), &(arot.z) );

						u = GetUserFromNum(dockedto);
						if( u ) {
		
							VECTOR3 gdockpos;
							u->v->dockedto = v->num;
							u->v->lastdock = t;
//							printf( "Receive dockpos(%d): %.2f %.2f %.2f\n", u->v->num, dockpos.x, dockpos.y, dockpos.z );
//							vecunrot( &dockpos, &(v->arot), &gdockpos );
							gdockpos = -dockpos;
							u->v->dockpos = gdockpos;
							printf( "Setting dockpos(%d): %.2f %.2f %.2f\n", u->v->num, gdockpos.x, gdockpos.y, gdockpos.z );
							u->v->dockarot = arot;
						}

					}
					
//					printf("R: '%s'\n", buf );
					if( strlen( buf ) > 2 && buf[0] == 'U' ) {
						int n;
						USER *u;
						n = atoi( buf+2 );
						char buf[128];
						u = GetUserFromNum(n);
						if( !u )
							break;
					
						u->v->pos = u->v->pos + v->dockpos ;
						v->pos = v->pos + u->v->dockpos ;

						(*j)->sendnetworkdata();
						u->sendnetworkdata();

						snprintf( buf, sizeof( buf )-1, "U %d\n", v->num );
						nsend( u, buf, strlen(buf) );

						u->v->dockedto = 0;
						u->v->lastdock = 0.0;


					}

				
				}

			}

		}
	}


}


void USER::sendnetworkdata() {

	char buf[256];
	char buf2[4096];
	std::vector<HVESSEL*>::iterator i;
	std::vector<BODY*>::iterator j;
	unsigned int n;

	netfreq = GetLagAvg();
	if( netfreq < 0.05 ) {
		netfreq = 0.05;
	}
	else if( netfreq > 2.0 ) {
		netfreq = 2.0;
	}

	if( t - lastgen < netfreq * TIMEACCEL ) {
//		printf("Vessel %d netfreq: %.2f\n", v->num, netfreq );
		/* updated too recently */
		return;
	}
	lastgen = t;
	
	if( fd == -1 )
		return;


	for( i = ves.begin() ; i < ves.end() ; i++ ) {
		VECTOR3 tmp;
		if( v->dockedto == (*i)->num && ( (*i)->num > v->num || (*i)->dockedto != v->num ) ) {
			if( v->lastdock+3.0*TIMEACCEL > t ) {
//				vecrot( &(v->dockpos), &((*i)->arot), &tmp );
				tmp = v->dockpos;
				v->pos = (*i)->pos - tmp;
				v->arot = v->dockarot;
				v->vel = (*i)->vel;
//				printf("Changing vessel %d to relative pos: %.2f %.2f %.2f\n", v->num, v->dockpos.x, v->dockpos.y, v->dockpos.z );
			}
		}
	}

	buf2[0] = 0;

	snprintf( buf, sizeof(buf)-1, "V %2.2i %f %f %f %f %f %f %f %f %f %f %2.2i\n",
		0, t, this->v->pos.x, this->v->pos.y, this->v->pos.z, this->v->vel.x, this->v->vel.y, this->v->vel.z,
		this->v->arot.x, this->v->arot.y, this->v->arot.z, 0 );
	strncat( buf2, buf, sizeof( buf2 )-strlen(buf2)-1 );


	for( i = ves.begin() ; i < ves.end() ; i++ ) {
		if( (*i)->num != this->v->num ) {
			snprintf( buf, sizeof(buf)-1, "V %2.2i %f %f %f %f %f %f %f %f %f %f %2.2i\n",
			( (*i) == v ? 0 : (*i)->num ), t, (*i)->pos.x, (*i)->pos.y, (*i)->pos.z, (*i)->vel.x, (*i)->vel.y, (*i)->vel.z,
			(*i)->arot.x, (*i)->arot.y, (*i)->arot.z, 0 );
			strncat( buf2, buf, sizeof( buf2 )-strlen(buf2)-1 );
		}
	}
	nsend( this, buf2, strlen(buf2) );

	buf2[0] = 0;
	for( j = bod.begin(), n=0 ; j < bod.end() ; j++, n++ ) {
		snprintf( buf, sizeof(buf)-1, "B %2.2i %f %f %f %f %f %f %f %f %2.2i\n",
		n, t, (*j)->pos.x, (*j)->pos.y, (*j)->pos.z, (*j)->vel.x, (*j)->vel.y, (*j)->vel.z,
		0.0, 0 );
		strncat( buf2, buf, sizeof( buf2 )-strlen(buf2)-1 );
	}
	nsend( this, buf2, strlen(buf2) );
	
}

void collisioncheck( VECTOR3 *p, VECTOR3 *v ) {
	
	VECTOR3 diff;
	double r;
		
	diff = *p; /* would normally be minus planet location */
	r = mod( diff );

	
	if( r < RADIUS ) {
		VECTOR3 zenith;
		VECTOR3 comp;
		double zdotv;
		zenith = ( *p / mod(p) );  // zenith is a unit vector pointing to the craft from the centre of the planet
		if( fabs( mod(&zenith) - 1.0 ) > 1e-7 ) { // check it's unity and bottle if not
			return;
		}
		
		zdotv = zenith.x*v->x + zenith.y*v->y + zenith.z*v->z; // the component of the velocity acting along the zenith vector

		comp = zenith * zdotv * 0.2; // that component in vector form, with a dampning factor to indicate inelasticity
		
		v->x += comp.x * -2.0; // take off that component and add it in the other direction
		v->y += comp.y * -2.0;
		v->z += comp.z * -2.0;
		
		*p = zenith * RADIUS; // ensure we don't fall below ground level
	}	

}

void edgecheck( VECTOR3 *pos ) {

	if( pos->x < WINX1 )
		pos->x = WINX2;
	else if( pos->x > WINX2 )
		pos->x = WINX1;

	if( pos->y < WINY1 )
		pos->y = WINY2;
	else if( pos->y > WINY2 )
		pos->y = WINY1;
	
}











/* MAIN */

int main(int argc, char *argv[]) {

	double oldt;
	double lastrep;
	
#ifndef _WIN32
	struct timeval tv;
	struct timezone tz;
#endif
	
	std::vector<USER*>::iterator j;
	
	_PORT = 1977;

	if( argc == 2 && atol(argv[1]) != 0 ) {
		_PORT = atol(argv[1]);
	}

	//	resetships();
	obj.clear();
	ves.clear();
	bod.clear();
	use.clear();
	
#ifdef _WIN32
	start = (float)timeGetTime()/1000.0;
#else
	gettimeofday(&tv,&tz);
	start = tv.tv_sec + (double)tv.tv_usec/1e6;
#endif
	
	oldt = 0.0;
	
	lastrep = -1.0 - REPORTSTEP;
	
	startnetworking();


	/* main loop */
	
	unsigned int robin = 0;
	
	for( ; ;  ) {

#ifdef _WIN32
		t = TIMEACCEL * ((float)timeGetTime()/1000.0 - start);
#else
		gettimeofday(&tv,&tz);
		t = ( tv.tv_sec + (double)tv.tv_usec/1e6 ) - start;
		t *= TIMEACCEL;
#endif
		dt = t - oldt;

		if( dt < 0.005 ) {
//			_sleep(5);
//			usleep(1);
		}
		else {
			
			oldt = t;


			std::vector<HVESSEL*>::iterator i;
			for( i = ves.begin() ; i < ves.end() ; i++ ) {
				physics4( &(*i)->pos, &(*i)->vel, dt, &(*i)->uacc );
//				collisioncheck( &(*i)->pos, &(*i)->vel );
//				edgecheck( &(*i)->pos );
				if( t - (*i)->lasertime > ( LASERTIME*TIMEACCEL ) )
					(*i)->laser = false;
			}

		}		

		int fdl;
		std::vector<USER*>::iterator user;
		for( user = use.begin() ; user < use.end() ; user++ ) { 
			fdl = (*user)->fd;
			if( fdl > 0 ) {
				(*user)->sendnetworkdata();
			}
		}
		
		serverlisten();

		for( j = use.begin() ; j < use.end() ; j++ ) {
		// old undocking -- not effective
/*			if( (*j)->v->dockedto && (*j)->v->lastdock + 3.0*TIMEACCEL < t ) {
				USER *ru;
				ru = GetUserFromNum((*j)->v->dockedto);
				if( ru )
					ru->v->dockedto = 0;
				(*j)->v->dockedto = 0;
			}
*/

			
			fdl = (*j)->fd;
			if( fdl == -1 ) {
				delete (*j);
			}
/*			else if( (*j)->ping + 3.0*TIMEACCEL < t ) {
				shutdown( (*j)->fd, 2 );
				(*j)->fd = -1;
				delete (*j);
			}
*/

		}
	}
	
	return 0;
	
}

