/*
	Project Hamac
	Copyright 2004/5 D Henderson dh219@aibs.org.uk

	Contributors:
	David Rowbotham
	Toni Ylisirnio
	Russell Smith

	Licenced under the GNU General Public Licence (GPL)
	http://www.gnu.org/licenses/gpl.txt

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "../defines.h"


unsigned int nextves() {
	
	std::vector<HVESSEL*>::iterator i;
	unsigned int n;
	bool nfree;
	
	
	for( n = 1, nfree = false ; n < MAXUSERS-1 && !nfree ; n++ ) {
		nfree = true;
		for( i = ves.begin() ; i < ves.end() ; i++ ) {
			if( (*i)->num == n )
				nfree = false;
		}
	}
	if( nfree == true)
		return n-1;
	
	return 999;
}


USER::USER( int filedesc, char *ipaddr, struct sockaddr_in *sockaddr ) {

	fd = filedesc;
	lastgen = 0.0;
	lastship = 0.0;

	ping=0.0;
	
	strncpy( ip, ipaddr, 20 );
	ip[19] = 0;

//	add.sin_addr.s_addr = 0;
	add = *sockaddr;

	udpport = -1;
	
	use.push_back(this);
	
	VECTOR3 pos,vel;
	pos = _V(2e6,0.0,0.0);
	vel = _V(0.0,sqrt(GM/2e6),0.0);
//	vel = _V(0.0,0.0,0.0);

	if( ves.size() > 0 ) {
		VECTOR3 tmp;
		tmp = _V(0.0,0.0,20.0);
		pos = (*ves.begin())->pos+tmp;
		vel = (*ves.begin())->vel;
	}
	
	userid = nextves();
	if( userid == 999 ) {
		return;
	}
	

	
	netfreq=0.3;
	
	int i;
	for( i = 0 ; i < 30 ; i++ )
		lag[i]=-1.0;
	lagno = 0;
	
	
	v = new HVESSEL( "User HVESSEL", 20e3, 10.0, &pos, &vel, this, userid );

}

USER::~USER() {

	delete v;

	std::vector<USER*>::iterator i;
	for( i = use.begin() ; i < use.end() ; i++ ) {
		if( *i == this ) {
			use.erase(i);
			break;
		}
	}

}

void USER::InsertLag(double lagtime) {

	if( lagno == 0 && lag[0] < 0.0 ) {
		lag[lagno++]=0.3;
		return;
	}

	lag[lagno++] = lagtime;
	if( lagno >= 30 )
		lagno=0;

}

double USER::GetLagAvg(){

	int i;
	int n;
	double sigma;

	sigma=0.0;
	n = 0;
	for( i = 0 ; i < 30 ; i++ ) {
		if( lag[i] > 0.0001 ) {
			sigma += lag[i];
			n++;
		}
	}

	if( n == 0 )
		return 0.3;

	return( sigma/(double)n );
}



const double radius[]= { 696000000.0, 1.0, 1.0, 6372795.0 };

VECTOR3 vacc;
double shiphead;
double oldshiphead;

struct ORIENT4 {
	VECTOR3 v;
	double rot;
} ;

struct ORIENT4 o;
struct ORIENT4 oldo;


void getaccel( VECTOR3 *accel, VECTOR3 *pos, VECTOR3 *uacc ) {
	
	double r2 = pos->x * pos->x + pos->y * pos->y + pos->z * pos->z;
	double accelmag = -1.0 * GM / r2;
		
	accel->x = accelmag * pos->x/sqrt(r2);
	accel->y = accelmag * pos->y/sqrt(r2);
	accel->z = accelmag * pos->z/sqrt(r2);

	accel->x += uacc->x;
	accel->y += uacc->y;
	accel->z += uacc->z;
	
//	printf( "User: %15.2f %15.2f %15.2f\n", uacc->x, uacc->y, uacc->z );
	
}


void physics4( VECTOR3 *pos, VECTOR3 *vel, double step, VECTOR3 *uacc ) {

	VECTOR3 pos1, pos2, pos3, pos4;
	VECTOR3 vel1, vel2, vel3, vel4;
	VECTOR3 acc1, acc2, acc3, acc4;
	
	double h = step;
	double hh = step / 2.0;
	
	/* k1 */

	pos1 = *pos;
	vel1 = *vel;
	getaccel( &acc1, &pos1, uacc );
	
	
	/* k2 */

	pos2 = pos1 + vel1 * hh;
	vel2 = vel1 + acc1 * hh;

	getaccel( &acc2, &pos2, uacc );
	
	/* k3 */
	
	pos3 = pos1 + vel2 * hh;
	vel3 = vel1 + acc2 * hh;
	
	getaccel( &acc3, &pos3, uacc );
	
	/* k4 */
	
	pos4 = pos1 + vel3 * h;
	vel3 = vel1 + acc3 * h;

	vel4 = vel1 + acc3 * h;
	
	getaccel( &acc4, &pos4, uacc );

	/****/
	
	pos->x += (h/6.0) * ( vel1.x + 2.0 * vel2.x + 2.0 * vel3.x + vel4.x );
	pos->y += (h/6.0) * ( vel1.y + 2.0 * vel2.y + 2.0 * vel3.y + vel4.y );
	pos->z += (h/6.0) * ( vel1.z + 2.0 * vel2.z + 2.0 * vel3.z + vel4.z );
	
	vel->x += (h/6.0) * ( acc1.x + 2.0 * acc2.x + 2.0 * acc3.x + acc4.x );
	vel->y += (h/6.0) * ( acc1.y + 2.0 * acc2.y + 2.0 * acc3.y + acc4.y );
	vel->z += (h/6.0) * ( acc1.z + 2.0 * acc2.z + 2.0 * acc3.z + acc4.z );
	
}

